/*
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.xwiki.extension.index.security;

import java.util.List;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.xwiki.stability.Unstable;
import org.xwiki.text.XWikiToStringBuilder;

import static java.util.Comparator.comparingDouble;

/**
 * Store the known security vulnerabilities for a given extension.
 *
 * @version $Id: 644b20199bf724be105f8e2f9372b77083c5def2 $
 * @since 15.5RC1
 */
@Unstable
public class ExtensionSecurityAnalysisResult
{
    private List<SecurityVulnerabilityDescriptor> securityVulnerabilities;

    private String advice;

    private boolean fromEnvironment;

    private boolean isInstalledExtension;

    /**
     * @param securityVulnerabilities the security vulnerabilities associated with the analyzed extension
     * @return the current object
     */
    public ExtensionSecurityAnalysisResult setResults(List<SecurityVulnerabilityDescriptor> securityVulnerabilities)
    {
        this.securityVulnerabilities = securityVulnerabilities;
        return this;
    }

    /**
     * @return the security vulnerabilities associated with the analyzed extension
     */
    public List<SecurityVulnerabilityDescriptor> getSecurityVulnerabilities()
    {
        return this.securityVulnerabilities;
    }

    /**
     * @return the translation key of the advice applicable on the security analysis (e.g., how to upgrade the extension
     *     to fix the identified security vulnerabilities)
     */
    public String getAdvice()
    {
        return this.advice;
    }

    /**
     * @return {@code true} when the extension is provided by the environment (e.g., from a servlet engine),
     *     {@code false} when the extension is from xwiki core, an installable extension, or their transitive
     *     dependencies
     * @since 15.6RC1
     */
    @Unstable
    public boolean isFromEnvironment()
    {
        return this.fromEnvironment;
    }

    /**
     * @param fromEnvironment {@code true} when the extension is provided by the environment (e.g., from a servlet
     *     engine), {@code false} when the extension is from xwiki core, an installable extension, or their transitive
     *     dependencies
     * @return the current object
     * @since 15.6RC1
     */
    @Unstable
    public ExtensionSecurityAnalysisResult setFromEnvironment(boolean fromEnvironment)
    {
        this.fromEnvironment = fromEnvironment;
        return this;
    }

    /**
     * @param advice the translation key of the advice applicable on the security analysis (e.g., how to upgrade the
     *     extension to fix the identified security vulnerabilities)
     * @return the current object
     */
    public ExtensionSecurityAnalysisResult setAdvice(String advice)
    {

        this.advice = advice;
        return this;
    }

    /**
     * @return the highest CVSS of this security vulnerabilities in this analysis
     */
    public Double getMaxCVSS()
    {
        if (this.securityVulnerabilities != null) {
            return this.securityVulnerabilities.stream()
                .max(comparingDouble(SecurityVulnerabilityDescriptor::getScore))
                .map(SecurityVulnerabilityDescriptor::getScore)
                .orElse(null);
        } else {
            return null;
        }
    }

    /**
     * @param isInstalledExtension {@code true} when the extension is installed, {@code false} otherwise
     * @since 15.6RC1
     */
    @Unstable
    public void setInstalledExtension(boolean isInstalledExtension)
    {
        this.isInstalledExtension = isInstalledExtension;
    }

    /**
     * @return {@code true} when the extension is installed, {@code false} otherwise
     * @since 15.6RC1
     */
    @Unstable
    public boolean isInstalledExtension()
    {
        return this.isInstalledExtension;
    }

    @Override
    public boolean equals(Object o)
    {
        if (this == o) {
            return true;
        }

        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        ExtensionSecurityAnalysisResult that = (ExtensionSecurityAnalysisResult) o;

        return new EqualsBuilder()
            .append(this.securityVulnerabilities, that.securityVulnerabilities)
            .append(this.advice, that.advice)
            .append(this.fromEnvironment, that.fromEnvironment)
            .append(this.isInstalledExtension, that.isInstalledExtension)
            .isEquals();
    }

    @Override
    public int hashCode()
    {
        return new HashCodeBuilder(17, 37)
            .append(this.securityVulnerabilities)
            .append(this.advice)
            .append(this.fromEnvironment)
            .append(this.isInstalledExtension)
            .toHashCode();
    }

    @Override
    public String toString()
    {
        return new XWikiToStringBuilder(this)
            .append("securityVulnerabilities", this.securityVulnerabilities)
            .append("advice", this.advice)
            .append("fromEnvironment", this.fromEnvironment)
            .append("isInstalledExtension", this.isInstalledExtension)
            .toString();
    }
}
