/*
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.xwiki.extension.security.internal.analyzer;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;

import javax.inject.Inject;
import javax.inject.Singleton;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.solr.client.solrj.SolrServerException;
import org.slf4j.Logger;
import org.xwiki.component.annotation.Component;
import org.xwiki.extension.CoreExtension;
import org.xwiki.extension.CoreExtensionFile;
import org.xwiki.extension.Extension;
import org.xwiki.extension.ExtensionFile;
import org.xwiki.extension.index.internal.ExtensionIndexStore;
import org.xwiki.extension.index.security.ExtensionSecurityAnalysisResult;
import org.xwiki.extension.index.security.SecurityVulnerabilityDescriptor;
import org.xwiki.extension.index.security.review.Review;
import org.xwiki.extension.index.security.review.ReviewResult;
import org.xwiki.extension.index.security.review.ReviewsMap;

import static org.apache.commons.lang.StringEscapeUtils.escapeXml;
import static org.apache.commons.lang3.exception.ExceptionUtils.getRootCauseMessage;
import static org.xwiki.extension.index.security.review.ReviewResult.UNSAFE;

/**
 * Update an {@link Extension} with the results of its latest {@link ExtensionSecurityAnalysisResult}.
 *
 * @version $Id: 26672a9f62e99367280a84a14dbb4c68786f85af $
 * @since 15.5RC1
 */
@Component(roles = VulnerabilityIndexer.class)
@Singleton
public class VulnerabilityIndexer
{
    /**
     * Shared constant when the suggestion is to upgrade the extension from the Extension Manager.
     */
    private static final String UPGRADE_FROM_EM_ADVICE = "extension.security.analysis.advice.upgradeFromEM";

    /**
     * Shared constant when the suggestion is to upgrade the extension XWiki itself.
     */
    private static final String UPGRADE_XWIKI_ADVICE = "extension.security.analysis.advice.upgradeXWiki";

    /**
     * Shared constant when the suggestion is to upgrade something else from the environment.
     */
    private static final String UPGRADE_ENVIRONMENT_ADVICE = "extension.security.analysis.advice.upgradeEnvironment";

    @Inject
    private ExtensionIndexStore extensionIndexStore;

    @Inject
    private Logger logger;

    /**
     * Update the security vulnerabilities known for the given extension.
     *
     * @param extension the extension to update
     * @param analysis the security analysis to update the extension
     * @param reviewsMap the map of reviewed CVEs
     * @return {@code true} if some new security vulnerabilities are inserted, {@code false} otherwise
     */
    public boolean update(Extension extension, ExtensionSecurityAnalysisResult analysis, ReviewsMap reviewsMap)
    {
        try {
            List<String> cveiDs = this.extensionIndexStore.getCVEIDs(extension.getId());

            List<SecurityVulnerabilityDescriptor> securityVulnerabilities = analysis.getSecurityVulnerabilities();
            boolean hasNew = securityVulnerabilities.stream()
                .anyMatch(vulnerability -> {
                    String vulnerabilityId = vulnerability.getId();
                    return vulnerability.getScore() > 0
                        && !cveiDs.contains(vulnerabilityId)
                        && isNotSafe(reviewsMap, vulnerabilityId);
                });

            boolean fromEnvironment = isFromEnvironment(extension);
            if (!securityVulnerabilities.isEmpty()) {
                if (fromEnvironment) {
                    analysis.setAdvice(UPGRADE_ENVIRONMENT_ADVICE);
                    analysis.setInstalledExtension(false);
                } else if (extension instanceof CoreExtension) {
                    analysis.setAdvice(UPGRADE_XWIKI_ADVICE);
                    analysis.setInstalledExtension(false);
                } else {
                    analysis.setAdvice(UPGRADE_FROM_EM_ADVICE);
                    analysis.setInstalledExtension(true);
                }
            }

            securityVulnerabilities.forEach(securityVulnerabilityDescriptor -> {
                Optional<List<Review>> byId = reviewsMap.getById(securityVulnerabilityDescriptor.getId());
                if (byId.isPresent()) {
                    securityVulnerabilityDescriptor.setReviews(formatReviews(byId.get()));
                } else {
                    securityVulnerabilityDescriptor.setReviews("");
                }
                securityVulnerabilityDescriptor
                    .setSafe(!isNotSafe(reviewsMap, securityVulnerabilityDescriptor.getId()));
            });

            this.extensionIndexStore.update(extension.getId(), analysis.setFromEnvironment(fromEnvironment));

            return hasNew;
        } catch (SolrServerException | IOException e) {
            this.logger.warn(
                "Failed to update the extension [{}] with the results of its latest security scan. Cause: [{}]",
                extension.getId(), getRootCauseMessage(e));
            return false;
        }
    }

    private static String formatReviews(List<Review> reviews)
    {
        String reviewsFormatted = reviews.stream()
            .filter(it -> it.getResult() == ReviewResult.SAFE)
            .map(VulnerabilityIndexer::formatReview).collect(Collectors.joining());
        boolean hasUnsafe = reviews.stream().anyMatch(review -> review.getResult() == UNSAFE);
        if (hasUnsafe) {
            String unsafeReviews = reviews
                .stream()
                .filter(review -> review.getResult() == UNSAFE)
                .map(VulnerabilityIndexer::formatReview)
                .collect(Collectors.joining());
            reviewsFormatted =
                String.format("<div class='box errormessage'>%s</div>%s", unsafeReviews, reviewsFormatted);
        }
        return reviewsFormatted;
    }

    private static boolean isNotSafe(ReviewsMap reviewsMap, String vulnerabilityId)
    {
        boolean isNotSafe;
        Map<String, List<Review>> map = reviewsMap.getReviewsMap();
        if (map.containsKey(vulnerabilityId)) {
            isNotSafe = map.get(vulnerabilityId)
                .stream()
                .anyMatch(review -> review.getResult() == UNSAFE);
        } else {
            isNotSafe = true;
        }
        return isNotSafe;
    }

    private boolean isFromEnvironment(Extension extension)
    {
        ExtensionFile file = extension.getFile();
        boolean isFromEnvironment;
        if (file instanceof CoreExtensionFile) {
            Path currentPathAbsolute = Path.of(".").toAbsolutePath();
            CoreExtensionFile coreExtensionFile = (CoreExtensionFile) file;
            if (coreExtensionFile.getURL() == null) {
                return false;
            }
            String withoutInternalReference = StringUtils.substringBefore(coreExtensionFile.getURL().toString(), "!")
                .replaceFirst("^jar:", "");
            try {
                Path jarPathAbsolute = FileUtils.toFile(new URL(withoutInternalReference)).toPath();
                isFromEnvironment = !currentPathAbsolute.relativize(jarPathAbsolute).startsWith(Path.of("webapps"));
            } catch (MalformedURLException e) {
                this.logger.warn("Failed to resolve the URL for [{}]. Cause: [{}]", coreExtensionFile,
                    getRootCauseMessage(e));
                isFromEnvironment = false;
            }
        } else {
            isFromEnvironment = false;
        }
        return isFromEnvironment;
    }

    private static String formatReview(Review review)
    {
        return String.format("<dl>\n"
                + "  <dt>%s</dt>\n"
                + "  <dd>%s</dd>\n"
                + "</dl>",
            escapeXml(review.getEmitter()),
            escapeXml(review.getExplanation()));
    }
}
